#include "window.h"
#include "screen.h"
#include "display.h"
#include <stdlib.h>
#include <assert.h>
#include <stdio.h>
#include <xcb/composite.h>
#include <xcb/damage.h>

typedef struct {
    /* public stuff */
    xcb_window_t     id;
    struct d_screen *sc;

    /* private stuff */
    int              ref;

    /* queried things, don't read them directly from the struct */
    int              x, y, w, h, bw;
    gboolean         mapped;
    gboolean         input_only;
    xcb_visualid_t   visual;
    xcb_pixmap_t     pixmap;

    gboolean         zombie;

    GHashTable      *plugin_data;

    xcb_damage_damage_t damage;

    xcb_get_window_attributes_cookie_t ck_get_attr;
    xcb_get_geometry_cookie_t          ck_get_geom;
    xcb_void_cookie_t                  ck_get_pixmap;
} d_window_priv_t;

static void window_get_attributes_reply(d_window_priv_t *w);
static void window_get_geometry_reply(d_window_priv_t *w);

d_window_t*
window_new(xcb_window_t id, struct d_screen *sc)
{
    d_window_priv_t *w;

    w = malloc(sizeof(d_window_priv_t));
    w->id = id;
    w->ref = 1;
    w->sc = sc;
    w->zombie = FALSE;
    w->pixmap = XCB_NONE;
    w->damage = XCB_NONE;

    screen_stacking_add(sc, (d_window_t*)w);

    w->ck_get_attr = xcb_get_window_attributes(sc->dpy->conn, id);
    w->ck_get_geom = xcb_get_geometry(sc->dpy->conn, id);

    w->ck_get_pixmap.sequence = 0;

    w->plugin_data = g_hash_table_new_full((GHashFunc)g_str_hash,
                                           (GEqualFunc)g_str_equal,
                                           g_free, NULL);

    //printf("new window 0x%x\n", w->id);

    return (d_window_t*)w;
}

void
window_ref(d_window_t *pubw)
{
    d_window_priv_t *w = (d_window_priv_t*)pubw;

    ++w->ref;
}

void
window_unref(d_window_t *pubw)
{
    d_window_priv_t *w = (d_window_priv_t*)pubw;

    if (w && --w->ref == 0) {
        xcb_pixmap_t p;

        screen_stacking_remove(w->sc, (d_window_t*)w);

        if ((p = window_get_pixmap(pubw))) {
            xcb_free_pixmap(w->sc->dpy->conn, p);
            w->pixmap = XCB_NONE;
        }

        g_hash_table_unref(w->plugin_data);
        free(w);
    }
}

static void
window_update_pixmap(d_window_priv_t *w)
{
    xcb_pixmap_t p;

    /* XXX can we save it for until we get the new pixmap? */
    if ((p = window_get_pixmap((d_window_t*)w))) {
        xcb_free_pixmap(w->sc->dpy->conn, p);
        w->pixmap = XCB_NONE;
    }

    //printf("updating pixmap for 0x%x\n", w->id);

    w->pixmap = xcb_generate_id(w->sc->dpy->conn);
    w->ck_get_pixmap = 
        xcb_composite_name_window_pixmap_checked(w->sc->dpy->conn,
                                                 w->id, w->pixmap);
    xcb_flush(w->sc->dpy->conn);
}

void
window_show(d_window_t *pubw)
{
    d_window_priv_t *w = (d_window_priv_t*)pubw;

    window_is_mapped(pubw); /* kill any ongoing request */

    //printf("show window 0x%x\n", w->id);

    window_update_pixmap(w);
    w->mapped = TRUE;
}

void
window_hide(d_window_t *pubw)
{
    d_window_priv_t *w = (d_window_priv_t*)pubw;

    window_is_mapped(pubw); /* kill any ongoing request */

    //printf("hide window 0x%x\n", w->id);

    w->mapped = FALSE;
}

void
window_become_zombie(d_window_t *pubw)
{
    d_window_priv_t *w = (d_window_priv_t*)pubw;

    if (w->zombie) return;

    w->zombie = TRUE;
}

gboolean
window_is_zombie(d_window_t *pubw)
{
    d_window_priv_t *w = (d_window_priv_t*)pubw;
    return w->zombie;
}

gboolean
window_is_input_only(d_window_t *pubw)
{
    d_window_priv_t *w = (d_window_priv_t*)pubw;
    if (w->ck_get_attr.sequence)
        window_get_attributes_reply(w);
    return w->input_only;
}

void
window_get_area(d_window_t *pubw, int *x, int *y, int *width, int *height,
                int *border_width)
{
    d_window_priv_t *w = (d_window_priv_t*)pubw;
    if (w->ck_get_geom.sequence)
        window_get_geometry_reply(w);
    *x = w->x;
    *y = w->y;
    *width = w->w;
    *height = w->h;
    *border_width = w->bw;
}

static void
window_get_attributes_reply(d_window_priv_t *w)
{
    xcb_get_window_attributes_reply_t *rep;
    xcb_generic_error_t *err = NULL;

    rep = xcb_get_window_attributes_reply(w->sc->dpy->conn,
                                          w->ck_get_attr,
                                          &err);

    if (rep) {
        w->input_only = rep->_class == XCB_WINDOW_CLASS_INPUT_ONLY;
        w->mapped = rep->map_state != XCB_MAP_STATE_UNMAPPED;
        w->visual = rep->visual;
        free(rep);
    }
    else {
        w->input_only = TRUE;
        w->mapped = FALSE;
        w->visual = XCB_NONE;
    }
    if (err) {
        printf("error getting attributes for window 0x%x\n", w->id);
        free(err);
    }
    w->ck_get_attr.sequence = 0;
}

static void
window_get_geometry_reply(d_window_priv_t *w)
{
    xcb_get_geometry_reply_t *rep;
    xcb_generic_error_t *err = NULL;

    rep = xcb_get_geometry_reply(w->sc->dpy->conn,
                                 w->ck_get_geom,
                                 &err);

    if (rep) {
        w->x = rep->x;
        w->y = rep->y;
        w->w = rep->width;
        w->h = rep->height;
        w->bw = rep->border_width;
        free(rep);
    }
    else {
        w->x = w->y = -1;
        w->w = w->h = 1;
        w->bw = 0;
    }
    if (err) {
        printf("error getting geometry for window 0x%x\n", w->id);
        free(err);
    }
    w->ck_get_geom.sequence = 0;
}

gboolean
window_is_mapped(d_window_t *pubw)
{
    d_window_priv_t *w = (d_window_priv_t*)pubw;
    if (w->ck_get_attr.sequence)
        window_get_attributes_reply(w);
    return w->mapped;
}

xcb_pixmap_t
window_get_pixmap(d_window_t *pubw)
{
    d_window_priv_t *w = (d_window_priv_t*)pubw;

    if (w->ck_get_pixmap.sequence) {
        xcb_generic_error_t *err;
        //printf("** checking get pixmap 0x%x\n", w->id);
        err = xcb_request_check(w->sc->dpy->conn, w->ck_get_pixmap);
        if (err) {
            w->pixmap = XCB_NONE;
            printf("error getting named pixmap for window 0x%x\n", w->id);
            free(err);
        }
        w->ck_get_pixmap.sequence = 0;
    }
    //printf("returning pixmap 0x%x for window 0x%x\n", w->pixmap, w->id);
    return w->pixmap;
}

xcb_visualid_t
window_get_visual(d_window_t *pubw)
{
    d_window_priv_t *w = (d_window_priv_t*)pubw;
    if (w->ck_get_attr.sequence)
        window_get_attributes_reply(w);
    return w->visual;
}

void
window_configure(d_window_t *pubw, int x, int y, int width, int height,
                 int border_width)
{
    d_window_priv_t *w = (d_window_priv_t*)pubw;

    /* this overrides any reply from our get_geometry call */
    if (w->ck_get_geom.sequence)
        w->ck_get_geom.sequence = 0;
    w->x = x;
    w->y = y;
    w->w = width;
    w->h = height;
    w->bw = border_width;
}

void
window_reconfigure(d_window_t *w)
{
    window_update_pixmap((d_window_priv_t*)w);
}

void
window_add_plugin_data(d_window_t *pubw, const char *key, void *data)
{
    d_window_priv_t *w = (d_window_priv_t*)pubw;
    char *skey = g_strdup(key);
    g_hash_table_replace(w->plugin_data, skey, data);
}

void*
window_find_plugin_data(d_window_t *pubw, const char *key)
{
    d_window_priv_t *w = (d_window_priv_t*)pubw;
    return g_hash_table_lookup(w->plugin_data, key);
}

void
window_remove_plugin_data(d_window_t *pubw, const char *key)
{
    d_window_priv_t *w = (d_window_priv_t*)pubw;
    g_hash_table_remove(w->plugin_data, key);
}

void
window_create_damage(d_window_t *pubw)
{
    d_window_priv_t *w = (d_window_priv_t*)pubw;

    if (!window_is_input_only(pubw)) {
        assert(w->damage == XCB_NONE);
        w->damage = xcb_generate_id(w->sc->dpy->conn);
        xcb_damage_create(w->sc->dpy->conn, w->damage, w->id,
                          XCB_DAMAGE_REPORT_LEVEL_NON_EMPTY);
    }
}

void window_destroy_damage(d_window_t *pubw)
{
    d_window_priv_t *w = (d_window_priv_t*)pubw;

    if (w->damage) {
        xcb_damage_destroy(w->sc->dpy->conn, w->damage);
        w->damage = XCB_NONE;
    }
}
